# -*- coding: utf-8 -*-
# models definitions for request_profiler
import datetime
import re

from django.contrib.auth.models import User
from django.db import models

from request_profiler import settings


class RuleSetManager(models.Manager):
    """Custom model manager for RuleSet instances."""
    def live_rules(self):
        """Return enabled rules.

        TODO: cache results.
        """
        return self.get_queryset().filter(enabled=True)


class RuleSet(models.Model):
    """Set of rules to match a URI and/or User."""

    enabled = models.BooleanField(
        default=True,
        db_index=True,
    )
    uri_regex = models.CharField(
        blank=True,
        default="",
        max_length=100,
        help_text=u"Regex used to filter by request URI."
    )
    user_group_filter = models.CharField(
        blank=True,
        default="",
        max_length=100,
        help_text=u"Name of a group used to filter users to profile.",
        verbose_name=u"User Group Filter"
    )
    include_anonymous = models.BooleanField(
        default=True,
        help_text=u"Include anonymous users (ignore group filters)."
    )
    # use the custom model manager
    objects = RuleSetManager()

    @property
    def has_group_filter(self):
        return len(self.user_group_filter.strip()) > 0

    def match_uri(self, request_uri):
        """Return True if there is a uri_regex and it matches.

        Args:
            request_uri: string, the HttpRequest.build_absolute_uri(), used
                to match against all the uri_regex.

        Returns True if there is a uri_regex and it matches, or if there
            there is no uri_regex, in which the match is implicit.

        """
        regex = self.uri_regex.strip()
        if regex == "":
            return True
        else:
            return re.search(regex, request_uri) is not None

    def match_user(self, user):
        """Return True if the user passes the various user filters.

        Match user against anonymous and group filters.
        """
        # if the user is anonymous, then they'll have no groups, and
        # the group filter is irrelevant - if you're including anonymous
        # that's all that counts
        if user is None or user.is_anonymous():
            return self.include_anonymous
        elif user.is_staff and settings.IGNORE_STAFF:
            return False

        # user is authenticated and not staff - do we have a group filter?
        if self.has_group_filter:
            # does the group exist in the user's groups
            return self.user_group_filter.strip() in [g.name for g in user.groups.all()]

        return True


class ProfilingRecord(models.Model):

    """Record of a request and its response."""
    user = models.ForeignKey(User, null=True, blank=True)
    session_key = models.CharField(blank=True, max_length=40)
    start_ts = models.DateTimeField(verbose_name="Request started at")
    end_ts = models.DateTimeField(verbose_name="Request ended at")
    duration = models.FloatField(verbose_name="Request duration (sec)")
    http_method = models.CharField(max_length=10)
    request_uri = models.URLField(verbose_name="Request path")
    remote_addr = models.CharField(max_length=100)
    http_user_agent = models.CharField(max_length=400)
    view_func_name = models.CharField(max_length=100, verbose_name="View function")
    response_status_code = models.IntegerField()

    def start(self):
        """Set start_ts from current datetime."""
        self.start_ts = datetime.datetime.utcnow()
        self.end_ts = None
        self.duration = None
        return self

    @property
    def elapsed(self):
        """Calculated time elapsed so far."""
        assert self.start_ts is not None, u"You must 'start' before you can get elapsed time."
        return (datetime.datetime.utcnow() - self.start_ts).total_seconds()

    def set_request_properties(self, request):
        """Extract values from HttpRequest and store locally."""
        self.http_method = request.method
        self.request_uri = request.path
        self.http_user_agent = request.META.get('HTTP_USER_AGENT'),
        # X-Forwarded-For is used by convention when passing through
        # load balancers etc., as the REMOTE_ADDR is rewritten in transit
        self.remote_addr = (
            request.META.get('X-Forwarded-For')
            if 'X-Forwarded-For' in request.META
            else request.META.get('REMOTE_ADDR')
        )
        # these two require middleware, so may not exist
        if hasattr(request, 'session'):
            self.session_key = request.session.session_key or ""
        # NB you can't store AnonymouseUsers, so don't bother trying
        if hasattr(request, 'user') and request.user.is_authenticated():
            self.user = request.user

    def stop(self):
        """Set end_ts and duration from current datetime."""
        assert self.start_ts is not None, u"You must 'start' before you can 'stop'"
        self.end_ts = datetime.datetime.utcnow()
        self.duration = (self.end_ts - self.start_ts).total_seconds()
        return self



# -*- coding: utf-8 -*-
# tests for the request_profiler app
import datetime

from django.test import TestCase, RequestFactory
from django.contrib.auth.models import User, AnonymousUser, Group

from request_profiler.middleware import ProfilingMiddleware
from request_profiler.models import RuleSet, RuleSetManager, ProfilingRecord
from request_profiler.signals import request_profile_complete
from request_profiler import settings


def dummy_view_func(request, **kwargs):
    "Fake function to pass into the process_view method."
    pass


class MockSession():
    def __init__(self, session_key):
        self.session_key = session_key


class MockResponse():
    def __init__(self, status_code):
        self.status_code = status_code
        self.values = {}

    def __len__(self):
        return len(self.values)

    def __getitem__(self, key):
        # if key is of invalid type or value, the list values will raise the error
        return self.values[key]

    def __setitem__(self, key, value):
        self.values[key] = value

    def __delitem__(self, key):
        del self.values[key]

    def __iter__(self):
        return iter(self.values)

    # def __reversed__(self):
    #     return FunctionalList(reversed(self.values))


class RuleSetManagerTests(TestCase):
    """Basic model manager method tests."""

    def setUp(self):
        pass

    def test_live_rules(self):

        r1 = RuleSet(uri_regex="", enabled=True)
        r1.save()
        self.assertEqual(RuleSet.objects.live_rules().count(), 1)

        r2 = RuleSet(uri_regex="", enabled=True)
        r2.save()
        self.assertEqual(RuleSet.objects.live_rules().count(), 2)

        r2.enabled = False
        r2.save()
        self.assertEqual(RuleSet.objects.live_rules().count(), 1)


class RuleSetModelTests(TestCase):
    """Basic model properrty and method tests."""

    def setUp(self):
        pass

    def test_default_properties(self):
        ruleset = RuleSet()
        props = [
            ('enabled', True),
            ('uri_regex', ""),
            ('user_group_filter', ""),
            ('include_anonymous', True),
        ]
        for p in props:
            self.assertEqual(getattr(ruleset, p[0]), p[1])
        self.assertIsInstance(RuleSet.objects, RuleSetManager)

    def test_has_group_filter(self):
        ruleset = RuleSet("")
        filters = (
            ("", False),
            (" ", False),
            ("test", True),
        )
        for f in filters:
            ruleset.user_group_filter = f[0]
            self.assertEqual(ruleset.has_group_filter, f[1])

    def test_match_uri(self):
        ruleset = RuleSet("")
        uri = "/test/"
        regexes = (
            ("", True),
            (" ", True),
            ("^/test", True),
            (".", True),
            ("/x", False)
        )

        for r in regexes:
            ruleset.uri_regex = r[0]
            self.assertEqual(ruleset.match_uri(uri), r[1])

    def test_match_user(self):
        ruleset = RuleSet("")
        self.assertFalse(ruleset.has_group_filter)
        self.assertTrue(ruleset.include_anonymous)

        # start with no user / anonymous
        self.assertTrue(ruleset.match_user(None))
        self.assertTrue(ruleset.match_user(AnonymousUser()))

        # now exclude anonymous
        ruleset.include_anonymous = False
        self.assertFalse(ruleset.match_user(None))
        self.assertFalse(ruleset.match_user(AnonymousUser()))

        # create a real user, but still no group filter
        bob = User.objects.create_user("Bob")
        self.assertFalse(bob.groups.exists())
        self.assertFalse(bob.is_staff)
        self.assertTrue(bob.is_authenticated())
        self.assertTrue(ruleset.match_user(bob))

        # now create the filter, and check bob no longer matches
        ruleset.user_group_filter = "test"
        test_group = Group(name="test")
        test_group.save()
        self.assertFalse(ruleset.match_user(bob))

        # add bob to the group, and check he now matches
        bob.groups.add(test_group)
        self.assertTrue(bob.groups.filter(name="test").exists())
        self.assertTrue(ruleset.match_user(bob))

        # now make him is_staff, and check that he's no longer allowed
        bob.is_staff = True
        self.assertTrue(settings.IGNORE_STAFF)
        self.assertFalse(ruleset.match_user(bob))

        # and finally, allow staff
        settings.IGNORE_STAFF = False
        self.assertTrue(ruleset.match_user(bob))


class ProfilingRecordModelTests(TestCase):
    """Basic model properrty and method tests."""

    def setUp(self):
        pass

    def test_default_properties(self):
        profile = ProfilingRecord()
        props = [
            ('user', None),
            ('session_key', ""),
            ('start_ts', None),
            ('end_ts', None),
            ('duration', None),
            ('http_method', ""),
            ('request_uri', ""),
            ('remote_addr', ""),
            ('http_user_agent', ""),
            ('view_func_name', ""),
            ('response_status_code', None),
        ]
        for p in props:
            self.assertEqual(getattr(profile, p[0]), p[1])

    def test_start(self):
        profile = ProfilingRecord().start()
        self.assertIsNotNone(profile.start_ts)
        self.assertIsNone(profile.end_ts)
        self.assertIsNone(profile.duration)
        # now check again to see that end and duration are cleared
        profile.end_ts = datetime.datetime.utcnow()
        profile.duration = 1
        profile.start()
        self.assertIsNotNone(profile.start_ts)
        self.assertIsNone(profile.end_ts)
        self.assertIsNone(profile.duration)

    def test_stop(self):
        profile = ProfilingRecord()
        self.assertRaises(AssertionError, profile.stop)
        profile.start().stop()
        self.assertIsNotNone(profile.start_ts)
        self.assertIsNotNone(profile.end_ts)
        self.assertIsNotNone(profile.duration)
        self.assertTrue(profile.duration > 0)

    def test_elapsed(self):
        profile = ProfilingRecord()
        with self.assertRaises(AssertionError):
            profile.elapsed
        profile.start()
        self.assertIsNotNone(profile.elapsed)
        self.assertIsNone(profile.end_ts)
        self.assertIsNone(profile.duration)

    def test_set_request_properties(self):

        factory = RequestFactory()
        request = factory.get("/test")
        request.META['HTTP_USER_AGENT'] = "test-browser"
        profile = ProfilingRecord()

        profile.set_request_properties(request)
        self.assertEqual(profile.http_method, request.method)
        self.assertEqual(profile.request_uri, request.path)
        # for some reason user-agent is a tuple - need to read specs!
        self.assertEqual(profile.http_user_agent, ("test-browser",))
        self.assertEqual(profile.session_key, "")
        self.assertEqual(profile.user, None)

        # test that we can set the session
        request.session = MockSession("test-session-key")
        profile = ProfilingRecord()
        profile.set_request_properties(request)
        self.assertEqual(profile.session_key, "test-session-key")

        # test that we can set the user
        request.user = User.objects.create_user("bob")
        profile = ProfilingRecord()
        profile.set_request_properties(request)
        self.assertEqual(profile.user, request.user)

        # but we do not save anonymous users
        request.user = AnonymousUser()
        profile = ProfilingRecord()
        profile.set_request_properties(request)
        self.assertEqual(profile.user, None)


class ProfilingMiddlewareTests(TestCase):

    def setUp(self):
        self.factory = RequestFactory()
        self.anon = AnonymousUser()
        self.bob = User.objects.create_user("bob")
        self.god = User.objects.create_superuser("god", "iamthelaw", "")
        self.test_group = Group(name="test")
        self.test_group.save()
        # remove any existing external signal listeners
        request_profile_complete.receivers = []

    def test_match_rules(self):
        # rule1 - to match all users
        r1 = RuleSet(enabled=True, include_anonymous=True)
        self.assertTrue(r1.match_user(self.anon))

        request = self.factory.get('/')
        request.user = self.anon
        self.assertTrue(r1.match_uri, request.path)

        profiler = ProfilingMiddleware()
        self.assertEqual(profiler.match_rules(request, [r1]), [r1])

        # now change the uri_regex so we no longer get a match
        r1.uri_regex = "^xyz$"
        self.assertEqual(profiler.match_rules(request, [r1]), [])

        # now change the user_groups so we no longer get a match
        request.user = self.bob
        r1.uri_regex = ""
        r1.user_group_filter = "test"
        self.assertEqual(profiler.match_rules(request, [r1]), [])
        # add bob to the group
        self.bob.groups.add(self.test_group)
        self.assertEqual(profiler.match_rules(request, [r1]), [r1])

    def test_process_request(self):
        request = self.factory.get('/')
        profiler = ProfilingMiddleware()
        profiler.process_request(request)
        # this implicitly checks that the profile is attached,
        # and that start() has been called.
        self.assertIsNotNone(request.profiler_record.elapsed)

    def test_process_view(self):
        request = self.factory.get('/')
        request.profiler_record = ProfilingRecord()
        profiler = ProfilingMiddleware()
        profiler.process_view(request, dummy_view_func, [], {})
        self.assertEqual(request.profiler_record.view_func_name, "dummy_view_func")

    def test_process_response(self):

        request = self.factory.get('/')
        profiler = ProfilingMiddleware()
        with self.assertRaises(AssertionError):
            profiler.process_response(request, None)

        # try no matching rules
        request.profiler_record = ProfilingRecord().start()
        response = profiler.process_response(request, MockResponse(200))
        self.assertEqual(response.status_code, 200)
        self.assertFalse(hasattr(request, 'profiler_record'))

        # try matching a rule, anc checking response values
        r1 = RuleSet(enabled=True, include_anonymous=True)
        r1.save()
        request.profiler_record = ProfilingRecord().start()
        response = profiler.process_response(request, MockResponse(200))
        self.assertTrue(request.profiler_record.response_status_code, response.status_code)
        self.assertTrue(response['X-Profiler-Duration'], request.profiler_record.duration)
        self.assertTrue(response['X-Profiler-Rules'], r1.id)

    def test_process_response_signal_cancellation(self):

        request = self.factory.get('/')
        request.profiler_record = ProfilingRecord().start()
        profiler = ProfilingMiddleware()

        # try matching a rule, anc checking response values
        r1 = RuleSet(enabled=True, include_anonymous=True)
        r1.save()

        self.signal_received = False

        def on_request_profile_complete(sender, **kwargs):
            self.signal_received = True
            return False

        request_profile_complete.connect(on_request_profile_complete)
        profiler.process_response(request, MockResponse(200))
        # because we returned False from the signal receiver,
        # we should have stopped profiling.
        self.assertTrue(self.signal_received)
        self.assertFalse(hasattr(request, 'profiler_record'))



# -*- coding: utf-8 -*-
# admin models for request_profiler app
from django.contrib import admin

from request_profiler import models


class RuleSetAdmin(admin.ModelAdmin):

    list_display = (
        'enabled',
        'uri_regex',
        'user_group_filter',
        'include_anonymous',
    )

admin.site.register(
    models.RuleSet,
    RuleSetAdmin
)


class ProfilingRecordAdmin(admin.ModelAdmin):

    list_display = (
        'user',
        'start_ts',
        'duration',
        'request_uri',
        'view_func_name',
        'response_status_code'
    )

admin.site.register(
    models.ProfilingRecord,
    ProfilingRecordAdmin
)



# -*- coding: utf-8 -*-
# signal definitions for request_profiler
from django.dispatch import Signal

# Signal sent after profile data has been captured, but before it is
# saved. If any receiving function returns False, the data is not saved.
# This can be used to, for instance, save the data asynchronously, by adding
# the `ProfilingRecord.save()` method call to a queue, and then returning False.
# The `instance` arg sent is a ProfilingRecord instance.
request_profile_complete = Signal(providing_args=['request', 'response', 'instance'])



# -*- coding: utf-8 -*-
# models definitions for request_profiler
from django.conf import settings

# default to ignoring all is_staff requests - otherwise you end up in a
# recursive nightmare when looking at the admin site.
IGNORE_STAFF = getattr(settings, 'REQUEST_PROFILER_IGNORE_STAFF', True)



# -*- coding: utf-8 -*-
import logging

from django.contrib.auth.models import AnonymousUser

from request_profiler.models import RuleSet, ProfilingRecord
from request_profiler.signals import request_profile_complete

logger = logging.getLogger(__name__)


class ProfilingMiddleware(object):
    """Middleware used to time request-response cycle.

    This middleware uses the `process_request` and `process_response`
    methods to both determine whether the request should be profiled
    at all, and then to extract various data for recording as part of
    the profile.

    The `process_request` method is used to start the profile timing; the
    `process_response` method is used to extract the relevant data fields,
    and to stop the profiler.

    """
    def match_rules(self, request, rules):
        """Return subset of a list of rules that match a request."""
        user = getattr(request, 'user', AnonymousUser())
        return [r for r in rules if r.match_uri(request.path) and r.match_user(user)]

    def process_request(self, request):
        """Start profiling."""
        request.profiler_record = ProfilingRecord().start()

    def process_view(self, request, view_func, view_args, view_kwargs):
        """Add view_func to the profiler info."""
        request.profiler_record.view_func_name = view_func.__name__

    def process_response(self, request, response):
        """Add response information and save the profiler record.

        By the time we get here, we've run all the middleware, the view_func
        has been called, and we've rendered the templates.

        This is the last chance to override the profiler and halt the saving
        of the profiler record instance. This is done by sending out a signal
        and aborting the save if any listeners respond False.

        """
        assert getattr(request, 'profiler_record', None) is not None, (
            u"Request has no profiler_record attached."
        )
        # see if we have any matching rules
        matching_rules = self.match_rules(request, RuleSet.objects.live_rules())

        # clean up after outselves
        if len(matching_rules) == 0:
            del request.profiler_record
            return response

        profiler = request.profiler_record
        profiler.set_request_properties(request)
        profiler.response_status_code = response.status_code
        # send signal so that receivers can intercept profiler
        signal_responses = request_profile_complete.send(
            sender=self.__class__,
            request=request,
            response=response,
            instance=profiler
        )
        # if any signal receivers have returned False, then do **not** save
        if all([s[1] for s in signal_responses]):
            profiler.stop().save()
            response['X-Profiler-Duration'] = profiler.duration or profiler.elapsed
            response['X-Profiler-Rules'] = ','.join(str(r.id) for r in matching_rules)
        else:
            del request.profiler_record

        return response
